# FLENG - asm part of aarch64 runtime

    .text

    .balign 4

    .global _fl_check_atom
    // x0=val
_fl_check_atom:
    and x8, x0, #3
    cmp x8, #2
    bne l1
    ret
l1:
    mov x1, x0
    mov x0, x19
    mov x2, #1          // FL_NOT_AN_ATOM
    b _fl_rt_error

    .global _fl_check_cell
    // x0=val
_fl_check_cell:
    tst x0, #3
    bne l2
    ret
l2:
    mov x1, x0
    mov x0, x19
    mov x2, #2          // FL_NOT_A_CELL
    b _fl_rt_error

    .global _fl_check_integer
    // x0=val
_fl_check_integer:
    tst x0, #1         // FL_INT_BIT
    beq l3
    ret
l3:
    mov x1, x0
    mov x0, x19
    mov x2, #3          // FL_NOT_AN_INT
    b _fl_rt_error

    .global _fl_check_list
    // x0=val
_fl_check_list:
    tst x0, #3
    bne l4
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x01000000 // FL_LIST_TAG
    cmp w8, w9
    bne l4
    ret
l4:
    mov x1, x0
    mov x0, x19
    mov x2, #4          // FL_NOT_A_LIST
    b _fl_rt_error

    .global _fl_check_module
    // x0=val
_fl_check_module:
    tst x0, #3
    bne l4b
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x0a000000 // FL_MODULE_TAG
    cmp w8, w9
    bne l4b
    ret
l4b:
    mov x1, x0
    mov x0, x19
    mov x2, #23          // FL_NOT_A_MDOULE
    b _fl_rt_error

    .global _fl_check_tuple
    // x0=val
_fl_check_tuple:
    tst x0, #3
    bne l5
    ldr w8, [x0]
    mov w9, #0x80000000  // FL_TUPLE_BIT
    tst w8, w9
    bne l5
    ret
l5:
    mov x1, x0
    mov x0, x19
    mov x2, #5          // FL_NOT_A_TUPLE
    b _fl_rt_error

    .global _fl_check_var
    // x0=val
_fl_check_var:
    tst x0, #3
    bne l6
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    beq l6
    mov w9, #0x07000000 // FL_REF_TAG
    cmp w8, w9
    beq l6
    mov x1, x0
    mov x0, x19
    mov x2, #6          // FL_NOT_A_VAR
    b _fl_rt_error
l6:
    ret

    .global _fl_check_number
    // x0=val
_fl_check_number:
    tst x0, #1
    beq l7
    ret
l7:
    tst x0, #2
    bne l8
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x02000000 // FL_FLOAT_TAG
    cmp w8, w9
    bne l8
    ret
l8:
    mov x1, x0
    mov x0, x19
    mov x2, #11          // FL_NOT_A_NUMBER
    b _fl_rt_error

    // x1=val -> d0
float_arg_1:
    tst x1, #1
    beq l9
    asr x1, x1, #1
    scvtf d0, x1
    ret
l9:
    ldr d0, [x1, #8]
    ret

    // x2=val -> d1
float_arg_2:
    tst x2, #1
    beq l10
    asr x2, x2, #1
    scvtf d1, x2
    ret
l10:
    ldr d1, [x2, #8]
    ret

    .global _fl_add
    // x1=num, x2=num -> x0
_fl_add:
    tst x1, #1
    beq l11
    tst x2, #1
    beq l11
    sub x0, x1, #1
    add x0, x0, x2
    ret
l11:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fadd d0, d0, d1
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_sub
    // x1=num, x2=num -> x0
_fl_sub:
    tst x1, #1
    beq l12
    tst x2, #1
    beq l12
    sub x1, x1, x2
    add x0, x1, #1
    ret
l12:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fsub d0, d0, d1
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_mul
    // x1=num, x2=num -> x0
_fl_mul:
    tst x1, #1
    beq l13
    tst x2, #1
    beq l13
    asr x1, x1, #1
    asr x2, x2, #1
    mul x0, x1, x2
    lsl x0, x0, #1
    orr x0, x0, #1
    ret
l13:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fmul d0, d0, d1
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_div
    // x1=num, x2=num -> x0
_fl_div:
    cmp x2, #1
    bne l14
    mov x0, x19
    mov x2, #12         // FL_DIV_BY_ZERO
    b _fl_rt_error
l14:
    tst x1, #1
    beq l15
    tst x2, #1
    beq l15
    asr x1, x1, #1
    asr x2, x2, #1
    sdiv x0, x1, x2
    lsl x0, x0, #1
    orr x0, x0, #1
    ret
l15:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    fdiv d0, d0, d1
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_mod
    // x1=num, x2=num -> x0
_fl_mod:
    cmp x2, #1
    bne l14b
    mov x0, x19
    mov x2, #12         // FL_DIV_BY_ZERO
    b _fl_rt_error
l14b:
    asr x1, x1, #1
    asr x2, x2, #1
    sdiv x0, x1, x2
    msub x0, x2, x0, x1
    lsl x0, x0, #1
    orr x0, x0, #1
    ret

    .global _fl_pow
    // x1=num, x2=num -> x0
_fl_pow:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    bl _pow
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_sqrt
    // x1=num -> x0
_fl_sqrt:
    str x30, [x28, #-8]!
    bl float_arg_1
    fsqrt d0, d0
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_sin
    // x1=num -> x0
_fl_sin:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _sin
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_cos
    // x1=num -> x0
_fl_cos:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _cos
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_tan
    // x1=num -> x0
_fl_tan:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _tan
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_atan
    // x1=num -> x0
_fl_atan:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _atan
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_loge
    // x1=num -> x0
_fl_loge:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _log
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_exp
    // x1=num -> x0
_fl_exp:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _exp
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_truncate
    // x1=num -> x0
_fl_truncate:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _trunc
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_round
    // x1=num -> x0
_fl_round:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _round
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_floor
    // x1=num -> x0
_fl_floor:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _floor
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_ceiling
    // x1=num -> x0
_fl_ceiling:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl _ceil
    ldr x30, [x28], #8
    b _fl_mkfloat

    .global _fl_shl
    // x1=num, x2=num -> x0
_fl_shl:
    lsr x2, x2, #1
    bic x1, x1, #1
    lsl x0, x1, x2
    orr x0, x0, #1
    ret

    .global _fl_shr
    // x1=num, x2=num -> x0
_fl_shr:
    lsr x2, x2, #1
    asr x0, x1, x2
    orr x0, x0, #1
    ret

    .global _fl_int
    // x1=num -> x0
_fl_int:
    tst x1, #1
    beq l16
    mov x0, x1
    ret
l16:
    ldr d0, [x1, #8]
    fcvtzs x0, d0
    lsl x0, x0, #1
    orr x0, x0, #1
    ret

    .global _fl_float
    // x1=num -> x0
_fl_float:
    asr x1, x1, #1
    scvtf d0, x1
    b _fl_mkfloat

true:
    adrp x0, _fl_true@PAGE
    add x0, x0, _fl_true@PAGEOFF
    ldr x0, [x0]
    ret

false:
    adrp x0, _fl_false@PAGE
    add x0, x0, _fl_false@PAGEOFF
    ldr x0, [x0]
    ret

    .global _fl_sametype
    // x1=val, x2=val -> x0
_fl_sametype:
    and x0, x1, #3
    and x8, x2, #3
    cmp x0, x8
    bne false
    tst x0, #3
    bne true
    ldr w0, [x1]
    ldr w8, [x2]
    mov w9, #0x80000000    // FL_TUPLE_BIT
    tst w0, w9
    beq l19
    tst w8, w9
    bne false
    b true
l19:
    tst w8, w9
    bne false
    mov w9, #0xff000000     // FL_TAG_MASK
    and w0, w0, w9
    and w8, w8, w9
    cmp w0, w8
    bne false
    b true

    .global _fl_equal
    // x1=val, x2=val -> x0
    // must preserve P0-P3
_fl_equal:
    cmp x1, x2
    beq true
    tst x1, #1
    beq l22
    tst x2, #1
    beq l23
    b false
l22:
    tst x2, #2
    bne false
    ldr w8, [x1]
    mov w9, #0xff000000     // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x02000000     // FL_FLOAT_TAG
    cmp w8, w9
    bne false
l23:
    tst x2, #2
    bne false
    tst x2, #1
    bne l25
    ldr w8, [x2]
    mov w9, #0xff000000     // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x02000000     // FL_FLOAT_TAG
    cmp w8, w9
    bne false
l25:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    fcmp d0, d1
    bne false
    b true

    .global _fl_less
    // x1=val, x2=val -> x0
    // must preserve P0-P3
_fl_less:
    tst x1, #1
    beq l26
    tst x2, #1
    beq l26
    cmp x1, x2
    bge false
    b true
l26:
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    fcmp d0, d1
    bmi true
    b false

first:
    mov x0, x1
    ret

second:
    mov x0, x2
    ret

    .global _fl_max
    // x1=val, x2=val -> x0
    // must preserve P0-P3
_fl_max:
    tst x1, #1
    beq l26a
    tst x2, #1
    beq l26a
    cmp x1, x2
    ble second
    b first
l26a:
    stp x1, x2, [x28, #-16]!
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    ldp x1, x2, [x28], #16
    fcmp d0, d1
    bmi second
    b first

    .global _fl_min
    // x1=val, x2=val -> x0
    // must preserve P0-P3
_fl_min:
    tst x1, #1
    beq l26i
    tst x2, #1
    beq l26i
    cmp x1, x2
    ble first
    b second
l26i:
    stp x1, x2, [x28, #-16]!
    str x30, [x28, #-8]!
    bl float_arg_1
    bl float_arg_2
    ldr x30, [x28], #8
    ldp x1, x2, [x28], #16
    fcmp d0, d1
    bmi first
    b second

    .global _fl_sign
    // x1=val -> x0
_fl_sign:
    tst x1, #1
    bne lsign1
    str x30, [x28, #-8]!
    bl float_arg_1
    ldr x30, [x28], #8
    adrp x0, zero@PAGE
    add x0, x0, zero@PAGEOFF
    ldr d1, [x0]
    fcmp d0, d1
    beq lsign3
    ble lsign2
lsign4:
    mov x0, #3
    ret
lsign2:
    mov x0, #-1
    ret
lsign3:
    mov x0, #1
    ret
lsign1:
    cmp x0, #1
    bmi lsign2
    bne lsign4
    mov x0, x1
    ret
zero: .double 0

    .global _fl_abs
    // x1=val -> x0
_fl_abs:
    tst x1, #1
    bne labs1
    str x30, [x28, #-8]!
    bl float_arg_1
    ldr x30, [x28], #8
    fabs d0, d0
    b _fl_mkfloat
labs1:
    cmp x1, #0
    bmi labs2
    mov x0, x1
    ret
labs2:
    neg x0, x1
    add x0, x0, #2
    ret

    .global _fl_rnd
    // x1=val -> x0
_fl_rnd:
    mov x0, x1
    b _fl_random_int

    .global _fl_fip
    // x1=val -> x0
_fl_fip:
    tst x1, #1
    beq lfip1
    mov x0, x1
    ret
lfip1:
    ldr x0, [x1, #8]
    lsr x0, x0, #52
    and x0, x0, #0x7ff
    sub x0, x0, #0x3ff
    cmp x0, #52
    blt lfip2
    mov x0, x1      // nan
    ret
lfip2:
    cmp x0, #0
    bge lfip4
    mov x0, #1
    ret
lfip4:
    mov x8, #-1
    lsr x8, x8, #12
    lsr x8, x8, x0
    ldr x0, [x1, #8]
    and x8, x0, x8
    cmp x8, #0
    bne lfip3
    mov x0, x1
    ret
lfip3:
    mvn x8, x8
    and x0, x8, x0
    fmov d0, x0
    b _fl_mkfloat

    .global _fl_ffp
    // x1=val -> x0
_fl_ffp:
    tst x1, #1
    beq lffp1
    mov x0, #1
    ret
lffp1:
    ldr x0, [x1, #8]
    lsr x0, x0, #52
    and x0, x0, #0x7ff
    sub x0, x0, #0x3ff
    cmp x0, #52
    blt lffp2
    cmp x0, #0x400
    bne lffp3
    ldr x0, [x1, #8]
    lsl x0, x0, #12
    cmp x0, #0
    beq lffp3
    mov x0, x1      // nan
    ret
lffp3:
    mov x0, #1
    ret
lffp2:
    cmp x0, #0
    bge lffp4
    mov x0, x1
    ret
lffp4:
    mov x8, #-1
    lsr x8, x8, #12
    lsr x8, x8, x0
    ldr x0, [x1, #8]
    and x8, x8, x0
    cmp x8, #0
    beq lffp3
    mvn x8, x8
    and x0, x0, x8
    fmov d1, x0
    ldr d0, [x1, #8]
    fsub d0, d0, d1
    b _fl_mkfloat

    .global _fl_below
    // x1=val, x2=val -> x0
_fl_below:
    sub x28, x28, #40
    str x1, [x28, #8]
    str x2, [x28, #16]
    str x3, [x28, #24]
    str x30, [x28, #32]
    mov x0, x19
    mov x3, x28
    bl _fl_ordering
    ldr x30, [x28, #32]
    ldr x3, [x28, #24]
    ldr x2, [x28, #16]
    ldr x1, [x28, #8]
    ldr x8, [x28]
    add x28, x28, #40
    tst x8, #1
    beq lbel1
    adrp x0, _fl_nil@PAGE
    add x0, x0, _fl_nil@PAGEOFF
    ldr x0, [x0]
    ret
lbel1:
    tst x0, #2     // -1
    bne true
    b false

    .global _fl_below_or_equal
    // x1=val, x2=val -> x0
_fl_below_or_equal:
    sub x28, x28, #40
    str x1, [x28, #8]
    str x2, [x28, #16]
    str x3, [x28, #24]
    str x30, [x28, #32]
    mov x0, x19
    mov x3, x28
    bl _fl_ordering
    ldr x30, [x28, #32]
    ldr x3, [x28, #24]
    ldr x2, [x28, #16]
    ldr x1, [x28, #8]
    ldr x8, [x28]
    add x28, x28, #40
    tst x8, #1
    beq lbeleq1
    adrp x0, _fl_nil@PAGE
    add x0, x0, _fl_nil@PAGEOFF
    ldr x0, [x0]
    ret
lbeleq1:
    cmp x0, #1
    beq false
    b true

    .global _fl_above
    // x1=val, x2=val -> x0
_fl_above:
    sub x28, x28, #40
    str x1, [x28, #8]
    str x2, [x28, #16]
    str x3, [x28, #24]
    str x30, [x28, #32]
    mov x0, x19
    mov x3, x28
    bl _fl_ordering
    ldr x30, [x28, #32]
    ldr x3, [x28, #24]
    ldr x2, [x28, #16]
    ldr x1, [x28, #8]
    ldr x8, [x28]
    add x28, x28, #40
    tst x8, #1
    beq labo1
    adrp x0, _fl_nil@PAGE
    add x0, x0, _fl_nil@PAGEOFF
    ldr x0, [x0]
    ret
labo1:
    cmp x0, #1
    bne false
    b true

    .global _fl_above_or_equal
    // x1=val, x2=val -> x0
_fl_above_or_equal:
    sub x28, x28, #40
    str x1, [x28, #8]
    str x2, [x28, #16]
    str x3, [x28, #24]
    str x30, [x28, #32]
    mov x0, x19
    mov x3, x28
    bl _fl_ordering
    ldr x30, [x28, #32]
    ldr x3, [x28, #24]
    ldr x2, [x28, #16]
    ldr x1, [x28, #8]
    ldr x8, [x28]
    add x28, x28, #40
    tst x8, #1
    beq laboeq1
    adrp x0, _fl_nil@PAGE
    add x0, x0, _fl_nil@PAGEOFF
    ldr x0, [x0]
    ret
laboeq1:
    cmp x0, #-1
    beq false
    b true

    .global _fl_deref
    // x0=val -> x0
    // must preserve P0-P3
_fl_deref:
    tst x0, #3
    beq l27
l28:
    ret
l27:
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    bne l28
    ldr x8, [x0, #8]
    cmp x0, x8
    beq l28
    mov x0, x8
    b _fl_deref

    .global _fl_unbox
    // x0=val -> x0
_fl_unbox:
    tst x0, #3
    beq lu27
lu28:
    ret
lu27:
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    bne lu29
    ldr x8, [x0, #8]
    cmp x0, x8
    beq l28
    mov x0, x8
    b _fl_unbox
lu29:
    mov x1, x0
    mov x0, x19
    b _fl_unbox_cell

    .global _fl_box
    // x1=val -> A
_fl_box:
    tst x1, #1
    bne box_l1
    ret
box_l1:
    mov x2, x30          // save return address
    adrp x0, _fl_nil@PAGE
    add x0, x0, _fl_nil@PAGEOFF
    ldr x0, [x0]
    mov x8, x0
    bl cons
    mov w9, #0x0f000000     // FL_BOX_LONG_TAG
    str w9, [x0]
    asr x1, x1, #1
    str x1, [x0, #8]
    ret x2

    .global _fl_unref
    // x8=val
    // must preserve A + P0-P3
_fl_unref:
    tst x8, #3
    beq l29
    ret
l29:
    mov w10, #0x00ffffff      // FL_COUNT_MASK
    ldr w11, [x8]
    and w11, w11, w10
    cmp w11, w10
    bne l30
    ret
l30:
    ldr w11, [x8]
    sub w11, w11, #1
    tst w11, w10
    beq l30a
    str w11, [x8]
    ret
l30a:
    sub x28, x28, #40
    stp x1, x2, [x28]
    stp x3, x30, [x28, #16]
    str x0, [x28, #32]
    mov x0, x19
    mov x1, x8
    bl _fl_release
    ldr x0, [x28, #32]
    ldp x3, x30, [x28, #16]
    ldp x1, x2, [x28]
    add x28, x28, #40
    ret

setcarry:
    mov x8, #1 << 29
    msr nzcv, x8
    ret

clrcarry:
    mrs x8, nzcv
    and x8, x8, #~(1 << 29)
    msr nzcv, x8
    ret

    .global _fl_force
    // x0=val -> x0, CF is 1 when suspending
_fl_force:
    tst x0, #3
    bne clrcarry
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x07000000 // FL_REF_TAG
    cmp w8, w9
    beq l33
    mov w9, #0x04000000 // FL_VAR_TAG
    cmp w8, w9
    bne clrcarry
    ldr x8, [x0, #8]
    cmp x0, x8
    beq l33
    mov x0, x8
    b _fl_force
l33:
    ldr x8, [x19, #8]   // tcb->sstackp
    str x0, [x8]
    ldr x0, [x20]      // goal->addr
    str x0, [x8, #8]
    add x8, x8, #16
    str x8, [x19, #8]   // tcb->sstackp
    b setcarry

    .global _fl_match
    // x1=val1, x2=val2 -> x0
    // assumes args are forced (allows initial quick eq check)
_fl_match:
    cmp x1, x2
    beq true
    mov x0, x19
    sub x28, x28, #32
    stp x1, x30, [x28]
    str x2, [x28, #16]
    bl _fl_match_rec
    ldr x2, [x28, #16]
    ldp x1, x30, [x28]
    add x28, x28, #32
    ret

    .global _fl_islist_t
    // x0=val, set CF if not a list
_fl_islist_t:
    tst x0, #3
    bne setcarry
    ldr w8, [x0]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w8, w8, w9
    mov w9, #0x01000000 // FL_LIST_TAG
    cmp w8, w9
    beq clrcarry
    b setcarry

    .global _fl_istuple_t
    // x0=val, set CF if not a tuple
_fl_istuple_t:
    tst x0, #3
    bne setcarry
    ldr w8, [x0]
    mov w9, #0x80000000 // FL_TUPLE_BIT
    tst w8, w9
    bne clrcarry
    b setcarry

    // x1=addr -> Carry (fail)
trynext:
    str x1, [x20]      // goal->addr
    b setcarry

    .global _fl_isval
    // x0=val, x2=atom, x1=addr
_fl_isval:
    cmp x0, x2
    bne trynext
    b clrcarry

    .global _fl_isfloat
    // x0=val, d0=float, x1=addr
_fl_isfloat:
    tst x0, #3
    bne trynext
    ldr d1, [x0, #8]
    fcmp d0, d1
    bne trynext
    b clrcarry

    .global _fl_islist
    // x0=val, x1=addr -> x0(car)
    // push cdr on stack
_fl_islist:
    str x30, [x28, #-8]!
    bl _fl_islist_t
    ldr x30, [x28], #8
    bcs trynext
    ldr x8, [x0, #16]
    str x8, [x28, #-8]!
    ldr x0, [x0, #8]
    b clrcarry

    .global _fl_istuple
    // x0=val, x1=addr, x2=arity -> x0(head)
    // push arglist on stack
_fl_istuple:
    tst x0, #3
    bne trynext
    ldr w8, [x0]
    mov w9, #0x80000000 // FL_TUPLE_BIT
    tst w8, w9
    beq trynext
    lsr w8, w8, #24
    and w8, w8, #0x7f
    cmp w8, w2
    bne trynext
    ldr x3, [x0, #16]
    str x3, [x28, #-8]!
    ldr x0, [x0, #8]
    b clrcarry

    // x8=car, x0=cdr -> x0
    // must preserve parameter registers
cons:
    ldr x9, [x19, #6 * 8]   // tcb->freelist
    cmp x9, #0
    bne l34
    sub x28, x28, #48
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    str x8, [x28, #32]
    str x0, [x28, #40]
    mov x0, x19
    bl _fl_new_chunk
    ldr x0, [x28, #40]
    ldr x8, [x28, #32]
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #48
    ldr x9, [x19, #6 * 8]   // tcb->freelist
l34:
    str x30, [x28, #-8]!
    ldr x10, [x9, #16]
    str x10, [x19, #6 * 8]  // tcb->freelist
    bl _fl_addref
    str x0, [x9, #16]
    mov x0, x8
    bl _fl_addref
    str x0, [x9, #8]
    ldr x10, [x19, #11 * 8]  // tcb->used
    add x10, x10, #1
    str x10, [x19, #11 * 8]    // tcb->used
    mov x0, x9
    ldr x30, [x28], #8
    ret

    .global _fl_mkfloat
    // d0=float64 -> x0
    // must preserve parameter registers
_fl_mkfloat:
    str x30, [x28, #-8]!
    mov x8, #1
    mov x0, #1
    bl cons
    str d0, [x0, #8]
    mov x8, #0x02000000      // FL_FLOAT_TAG
    str w8, [x0]
    ldr x30, [x28], #8
    ret

    .global _fl_mklist
    // x0=cdr, <stack>=car -> x0
_fl_mklist:
    ldr x8, [x28]
    str x30, [x28]
    bl cons
    mov x8, #0x01000000      // FL_LIST_TAG
    str w8, [x0]
    ldr x30, [x28], #8
    ret

    .global _fl_mktuple
    // x0=args, x8=arity, x9=atom -> x0
_fl_mktuple:
    stp x8, x30, [x28, #-16]!
    mov x8, x9
    bl cons
    ldp x8, x30, [x28], #16
    // fall through ...

    .global _fl_mktuple0
    // x0=args, x8=arity -> x0
_fl_mktuple0:
    lsl x8, x8, #24
    mov w9, #0x80000000 // FL_TUPLE_BIT
    orr w8, w8, w9
    str w8, [x0]
    ret

    .global _fl_mkvar
    // -> x0
_fl_mkvar:
    str x30, [x28, #-8]!
    mov x8, #1
    adrp x0, _fl_nil@PAGE
    add x0, x0, _fl_nil@PAGEOFF
    ldr x0, [x0]
    bl cons
    str x0, [x0, #8]
    mov x8, #0x04000000   // FL_VAR_TAG
    str w8, [x0]
    ldr x30, [x28], #8
    ret

    // x21=goal
enqueue:
    ldr x0, [x19, #13 * 8]     // tcb->qend
    str x21, [x0]
    add x0, x0, #8
    ldr x3, [x19, #12 * 8]     // tcb->qstart
    cmp x0, x3
    bne l40
    adrp x1, _fl_nil@PAGE
    add x1, x1, _fl_nil@PAGEOFF
    ldr x1, [x1]
    mov x2, #7                  // FL_QUEUE_FULL
    mov x0, x19
    b _fl_rt_error
l40:
    ldr x8, [x19, #15 * 8]     // tcb->max_goals
    lsl x8, x8, #3
    ldr x9, [x19, #14 * 8]     // tcb->queue
    add x8, x8, x9
    cmp x0, x8
    blo l41
    ldr x0, [x19, #14 * 8]     // tcb->queue
l41:
    str x0, [x19, #13 * 8]     // tcb->qend
    ldr x8, [x19, #10 * 8]     // tcb->active
    add x8, x8, #1
    str x8, [x19, #10 * 8]    // tcb->active
    ret

    .global _fl_opengoal
    // -> x21
_fl_opengoal:
    ldr x21, [x19, #16 * 8]     // tcb->freegoals
    cmp x21, #0
    bne lg1
    mov x2, #26                  // FL_TOO_MANY_GOALS
    mov x0, x19
    b _fl_rt_error
lg1:
    ldr x8, [x21]               // goal->addr
    ldr x0, [x20, #6 * 8]       // goal->task
    str x30, [x28, #-8]!
    bl _fl_addref
    ldr x30, [x28], #8
    str x0, [x21, #6 * 8]       // goal->task
    str x8, [x19, #16 * 8]      // tcb->freegoals
    ldr x8, [x19, #21 * 8]    // tcb->goals
    add x8, x8, #1
    str x8, [x19, #21 * 8]    // tcb->goals
    b enqueue

    .global _fl_dropgoal
_fl_dropgoal:
    ldr x8, [x19]               // tcb->goal
    ldr x9, [x19, #16 * 8]     // tcb->freegoals
    str x9, [x8]               // goal->addr
    str x8, [x19, #16 * 8]      // tcb->freegoals
    ldr x9, [x19, #21 * 8]    // tcb->goals
    sub x9, x9, #1
    str x9, [x19, #21 * 8]    // tcb->goals
    mov x1, #0
    str x1, [x8, #5 * 8]     // goal->info
    adrp x1, _fl_nil @PAGE
    add x1, x1, _fl_nil@PAGEOFF
    ldr x1, [x1]
    ldr x0, [x8, #6 * 8]      // goal->task
    str x1, [x8, #6 * 8]      // goal->task
    mov x8, x0
    b _fl_unref

    .global _fl_addref
    // x0=val
    // may not change any registers
_fl_addref:
    tst x0, #3
    beq l42
    ret
l42:
    stp x8, x9, [x28, #-16]!
    mov w9, #0x00ffffff     // FL_COUNT_MASK
    ldr w8, [x0]
    and w8, w8, w9
    cmp w8, w9
    beq l42a
    ldr w8, [x0]
    add w8, w8, #1
    str w8, [x0]
l42a:
    ldp x8, x9, [x28], #16
    ret

    .global _fl_puttailarg
    // x0=list
_fl_puttailarg:
    ldr x8, [x0, #16]
    tst x8, #3
    beq l43
    ldr x0, [x0, #8]
l43:
    str x30, [x28, #-8]!
    bl _fl_addref
    str x0, [x21, #4 * 8]   // goal->args[ 3 ]
    ldr x30, [x28], #8
    ret

    .global _fl_reify
    // replace cdr of pair with deref'd element
    // x0=pair
_fl_reify:
    str x30, [x28, #-8]!
    bl _fl_deref
    ldr x8, [x0, #16]
    stp x0, x8, [x28, #-16]!
    mov x0, x8
    bl _fl_deref
    bl _fl_addref
    ldp x1, x8, [x28], #16
    bl _fl_unref
    str x0, [x1, #16]
    ldr x30, [x28], #8
    ret

    .global _fl_resolve
    // x4=cache, x1=arity, x2=module, x3=name
_fl_resolve:
    ldr x8, [x4]
    cmp x8, #0
    beq l44
    str x8, [x21]        // goal->addr
    ret
l44:
    mov x0, x19
    mov x5, x21
    b _fl_resolve_pdef

    .global _fl_replace
    // x0=val, x2=argaddr
_fl_replace:
    ldr x8, [x2]
    cmp x0, x8
    bne l45
    ret
l45:
    ldr x8, [x2]
    str x30, [x28, #-8]!
    bl _fl_addref
    ldr x30, [x28], #8
    str x0, [x2]
    b _fl_unref

    .global _fl_tailgoal
    // x2=addr -> x21(N)
_fl_tailgoal:
    ldr x3, [x19, #16] // tcb->tailcalls
    ldr x1, [x19, #24] // tcb->timeslice
    cmp x3, x1
    blt l_tg_1
    stp x2, x30, [x28, #-16]!
    bl _fl_opengoal
    ldp x2, x30, [x28], #16
    str x2, [x21]      // goal->addr
    ret
l_tg_1:
    mov x21, x20
    str x2, [x21]      // goal->addr
    ret

    .global _fl_tail_jump
_fl_tail_jump:
    ldr x3, [x19, #16] // tcb->tailcalls
    ldr x1, [x19, #24] // tcb->timeslice
    cmp x3, x1
    blt l46
    mov x0, #0
    str x0, [x19, #16] // tcb->tailcalls
    ret
l46:
    add x3, x3, #1
    str x3, [x19, #16] // tcb->tailcalls
    // fall through ...

    .global _fl_invoke_jump
_fl_invoke_jump:
    mov x28, x29
    mov x20, x21
    str x21, [x19]      // tcb->goal
    ldr x0, [x21]      // goal->addr
    br x0

    .global _fl_test_integer
    // x1=val -> x0 (bool)
_fl_test_integer:
    tst x1, #1
    bne true
    b false

    .global _fl_test_number
    // x1=val -> x0 (bool)
_fl_test_number:
    tst x1, #1
    bne true
    mov w8, #0x02000000      // FL_FLOAT_TAG
    b _fl_test_tag

    .global _fl_test_atom
    // x1=val -> x0 (bool)
_fl_test_atom:
    mov x0, x1
    and x0, x0, #3
    cmp x0, #2
    bne false
    b true

    .global _fl_test_tag
    // x1=val, w8=tag -> x0 (bool)
_fl_test_tag:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w9, #0xff000000 // FL_TAG_MASK
    and w0, w0, w9
    cmp w0, w8
    bne false
    b true

    .global _fl_test_var
    // x1=val -> x0 (bool)
_fl_test_var:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w8, #0xff000000 // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x07000000 // FL_REF_TAG
    cmp w0, w8
    beq true
    mov w8, #0x04000000 // FL_VAR_TAG
    cmp w0, w8
    bne false
    ldr x0, [x1, #8]
    cmp x0, x1
    bne false
    b true

    .global _fl_test_remote
    // x1=val -> x0 (bool)
_fl_test_remote:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w8, #0xff000000 // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x07000000 // FL_REF_TAG
    cmp w0, w8
    beq true
    mov w8, #0x05000000 // FL_PORT_TAG
    cmp w0, w8
    bne false
    ldr x0, [x1, #8]
    tst x0, #1
    bne true
    b false

    .global _fl_test_tuple
    // x1=val -> x0 (bool)
_fl_test_tuple:
    tst x1, #3
    bne false
    ldr w0, [x1]
    mov w8, #0x80000000 // FL_TUPLE_BIT
    tst w0, w8
    beq false
    b true

    .global _fl_test_struct
    // x1=val, x2=name, x3=arity -> x0 (bool)
_fl_test_struct:
    tst x1, #3         // FL_BITS_MASK
    bne false
    ldr w0, [x1]
    lsr w5, w0, #24
    and w5, w5, #0xff
    lsr w3, w3, #1
    orr w3, w3, #0x80	 // FL_TUPLE_BIT
    cmp w5, w3
    bne false
    ldr x0, [x1, #8]
    cmp x0, x2
    bne false
    b true

    .global _fl_test_array
    // x1=val -> x0 (bool)
_fl_test_array:
    tst x1, #3         // FL_BITS_MASK
    bne false
    ldr w0, [x1]
    mov w8, #0xff000000     // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x08000000     // FL_ARRAY_TAG
    cmp w0, w8
    bne false
    b true

    .global _fl_enter
    // x0=tcb
_fl_enter:
    mov x19, x0
    ldr x28, [x19, #5 * 8]  // tcb->stackbase
    ldr x0, [x19, #22*8]    // tcb->sp0
    mov sp, x0
    mov x29, x28
    ldr x20, [x19]      // tcb->goal
    ldr x0, [x20]      // goal->addr
    br x0

    .global _fl_atomic
    // x1=val -> x0 (bool)
_fl_atomic:
    tst x1, #3         // FL_BITS_MASK
    bne true
    ldr w0, [x1]
    mov w8, #0xff000000     // FL_TAG_MASK
    and w0, w0, w8
    mov w8, #0x02000000     // FL_FLOAT_TAG
    cmp w0, w8
    bne false
    b true

    .global _fl_entry
    // x1=arity
_fl_entry:
    adrp x8, _fl_nil@PAGE
    add x8, x8, _fl_nil@PAGEOFF
    ldr x8, [x8]
    ldr x0, [x20, #6 * 8]    // goal->task
    cmp x0, x8
    beq l50a
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
    cmp x0, x8
    bne l50
l50a:
    ldr x0, [x19, #20 * 8]     // tcb->logging
    tst x0, #1
    bne l50
    ret
l50:
    mov x0, x19
    b _fl_log_entry

    .global _fl_assign0
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
_fl_assign0:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_assign
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_d_assign0
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
_fl_d_assign0:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_d_assign
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_unify
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
_fl_unify:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_unify_rec
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_d_unify
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
_fl_d_unify:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_d_unify_rec
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_unify_safe
    // x1=val, x2=val -> x0 (bool)
    // must preserve P0-P3
_fl_unify_safe:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_unify_safe1
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_unify_checked
    // x1=val, x2=val
    // must preserve P0-P3
_fl_unify_checked:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_unify_rec_checked
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_d_unify_checked
    // x1=val, x2=val
    // must preserve P0-P3
_fl_d_unify_checked:
    sub x28, x28, #32
    str x1, [x28]
    str x2, [x28, #8]
    str x3, [x28, #16]
    str x30, [x28, #24]
    mov x0, x19
    bl _fl_d_unify_rec_checked
    ldr x30, [x28, #24]
    ldr x3, [x28, #16]
    ldr x2, [x28, #8]
    ldr x1, [x28]
    add x28, x28, #32
    ret

    .global _fl_lookup
    // A, X0: table [count, int, label, ..., label]
_fl_lookup:
    mov x1, x8
    ldr x9, [x8], #8
llok1:
    ldr x5, [x8]
    cmp x0, x5
    beq llok2
    add x8, x8, #16
    subs x9, x9, #1
    bne llok1
    sub x8, x8, #8
llok2:
    ldr x8, [x8, #8]
    add x8, x1, x8
    br x8

    .global _fl_goto
    // A, X0: table [min, max*, flabel, label1, ...]
    // max: mark bit is 0
_fl_goto:
    mov x1, x8
    ldr x9, [x8]
    sub x0, x0, x9
    ldr x9, [x8, #8]
    cmp x0, x9
    bhi lgoto1
    add x0, x0, #6
    lsl x0, x0, #2
    add x0, x0, x1
    ldr x0, [x0]
    add x1, x0, x1
    br x1
lgoto1:
    ldr x0, [x8, #16]
    add x1, x0, x1
    br x1

    .global _fl_prepare_c_arg
    // A: Arg
_fl_prepare_c_arg:
    tst x0, #2      // FL_ATOM_BIT
    beq lprep1
    add x0, x0, #1
    ret
lprep1:
    ldr w8, [x0]
    and w8, w8, #0xff000000 // FL_TAG_MASK
    mov w9, #0x08000000 // FL_ARRAY_TAG
    cmp w8, w9
    bne lprep2
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
    ret
lprep2:
    add x0, x0, #8
    ret

    .global _fl_new_task
    // G, A = status(var)
_fl_new_task:
    mov x2, x30         // save return addr in P1
    mov x1, x0          // save status in P0
    adrp x8, _fl_nil@PAGE
    add x8, x8, _fl_nil@PAGEOFF
    ldr x8, [x8]
    ldr x0, [x20, #6*8]   // goal->task
    cmp x0, x8
    bne new_task_l1
    // cons new env/parent cell [[]|[]]]
    bl cons
    mov x9, #0x01000000   // FL_LIST_TAG
    str w9, [x0]
    mov x8, x0
    b new_task_l2
new_task_l1:
    ldr x0, [x0, #8]
    ldr x8, [x0, #8]            // env
new_task_l2:
    ldr x0, [x20, #6*8]            // goal->task (parent)
    bl cons
    mov x8, x0
    mov x0, x1
    bl cons                       // task: [env/parent|st]
    mov x9, #0x06000000        // FL_TASK_TAG
    orr x9, x9, #1
    str w9, [x0]
    ldr x8, [x20, #6*8]            // goal->task (old)
    bl _fl_unref
    str x0, [x20, #6*8]            // goal->task
    ret x2

    .global _fl_box_long_ref
    // A = box -> A
_fl_box_long_ref:
    ldr w8, [x0]
    and w8, w8, #0xff000000 // FL_TAG_MASK
    mov w9, #0x08000000 // FL_ARRAY_TAG
    cmp w8, w9
    bne lblr1
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
lblr1:
    ldr x0, [x0, #8]
fix:
    lsl x0, x0, #1
    orr x0, x0, #1
    ret

    .global _fl_box_int_ref
    // A = box -> A
_fl_box_int_ref:
    ldr w8, [x0]
    and w8, w8, #0xff000000 // FL_TAG_MASK
    mov w9, #0x08000000 // FL_ARRAY_TAG
    cmp w8, w9
    bne lbir1
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
lbir1:
    mov x8, #0
    ldr w8, [x0, #8]
    mov x0, x8
    b fix

    .global _fl_box_char_ref
    // A = box -> A
_fl_box_char_ref:
    ldr w8, [x0]
    and w8, w8, #0xff000000 // FL_TAG_MASK
    mov w9, #0x08000000 // FL_ARRAY_TAG
    cmp w8, w9
    bne lbcr1
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
lbcr1:
    ldrb w0, [x0, #8]
    b fix

    .global _fl_box_short_ref
    // A = box -> A
_fl_box_short_ref:
    ldr w8, [x0]
    and w8, w8, #0xff000000 // FL_TAG_MASK
    mov w9, #0x08000000 // FL_ARRAY_TAG
    cmp w8, w9
    bne lbsr1
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
lbsr1:
    ldrsh x0, [x0, #8]
    b fix

    .global _fl_box_double_ref
    // T, A = box -> A
_fl_box_double_ref:
    ldr w8, [x0]
    and w8, w8, #0xff000000 // FL_TAG_MASK
    mov w9, #0x08000000 // FL_ARRAY_TAG
    cmp w8, w9
    bne lbdr1
    ldr x0, [x0, #8]
    ldr x0, [x0, #8]
lbdr1:
    ldr d0, [x0, #8]
    b _fl_mkfloat
